"""Helper functions to make your life simple."""

from os import walk
from pathlib import Path

import aws_cdk as cdk
import yaml

from aws_cdk import Stack, Stage


def check_ansible_dir(directory: str) -> bool:
    """Checks if an Ansible directory exists.

    Parameters:
      - directory (str): The directory path to check.

    Returns:
      - bool: True if the directory exists, False otherwise.

    Functionality:
      - Resolves the full path using the __file__ variable.
      - Checks if the path exists and is a directory using pathlib.
      - Returns True if it's a directory, False otherwise.
    """

    this_dir = Path(__file__).parent
    ansible_path = this_dir.joinpath(directory).resolve()
    return ansible_path.is_dir()


def apply_tags(props: dict, resource: cdk.Stack | cdk.Stage) -> Stack | Stage:
    """Applies tags to a CDK Stack or Stage.

    Parameters:
      - props (dict): Dictionary containing tags to apply
      - resource (cdk.Stack | cdk.Stage): Stack or Stage to apply tags to

    Returns:
      - Stack | Stage: The same resource passed in, after tags applied

    Functionality:
      - Loops through tags in props and adds them to the resource with cdk.Tags.of
      - Returns the modified resource object
    """

    for key, value in props["tags"].items():
        cdk.Tags.of(resource).add(key, value)

    return resource


def _load_base_properties(stage: str) -> dict:
    """Loads base properties from the config-ci-cd.yaml file."""
    config_file_path = Path("cdk/config/config-ci-cd.yaml")
    with config_file_path.open(encoding="utf-8") as file:
        props = yaml.safe_load(file)
        props["stage"] = stage
    return props


def _load_stage_properties(stage: str) -> dict:
    """Loads stage-specific properties from files in the stage directory."""
    props_env: dict[str, str] = {}
    for dir_path, _, files in walk(f"cdk/config/{stage}", topdown=False):
        for file_name in files:
            file_path = Path(f"{dir_path}/{file_name}")
            with file_path.open(encoding="utf-8") as f:
                props_env.update(yaml.safe_load(f) or {})
    return props_env


def load_properties(stage: str) -> dict:
    """Loads properties from YAML configuration files.

    Args:
        stage (str): The deployment stage (e.g., 'dev', 'prod').

    Returns:
        dict: A dictionary containing the loaded properties.
    """

    props = _load_base_properties(stage)
    props_env = _load_stage_properties(stage)

    # Merge tags, giving precedence to stage-specific tags
    props_tags = props.get("tags", {})
    conf_tags = props_env.get("tags", {})
    updated_props = {
        **props_env,
        **props,
        "tags": {**props_tags, **conf_tags},
    }

    # Add optional properties if they exist in stage-specific config
    optional_keys = ["slack_channel_id_alarms", "ms_teams_channel_id_alarms"]
    for key in optional_keys:
        value = props_env.get(key)
        if value:
            updated_props[key] = value

    return updated_props



import aws_cdk as cdk
import aws_cdk.aws_ssm as ssm

from aws_cdk import Aspects
from cdk_nag import AwsSolutionsChecks
from cdk_opinionated_constructs.schemas.configuration_vars import ConfigurationVars
from cdk_opinionated_constructs.utils import load_properties
from constructs import Construct


class PipelinePluginsStack(cdk.Stack):
    def __init__(self, scope: Construct, construct_id: str, env, props, **kwargs) -> None:
        """Initializes the PipelinePluginsStack construct.

        Parameters:
        - scope (Construct): The parent construct.
        - construct_id (str): The construct ID.
        - env: The CDK environment.
        - props: Stack configuration properties.
        - **kwargs: Additional keyword arguments passed to the Stack constructor.

        The constructor does the following:

        1. Call the parent Stack constructor.

        2. Loads configuration from YAML files in the config directory for the stage.

        3. Merge the loaded configuration with the passed props.

        4. Create ConfigurationVars and PipelineVars objects from the configuration.

        5. Create an SSM StringParameter to store the pipeline plugins configuration
           from PipelineVars, for later retrieval.

        6. Validates the stack against the AWS Solutions checklist using Aspects.
        """

        super().__init__(scope, construct_id, env=env, **kwargs)
        config_vars = ConfigurationVars(**props)
        props_env = load_properties(stage=config_vars.stage)

        config_vars = ConfigurationVars(**props_env)

        ssm.StringParameter(
            self,
            id="pipeline_plugins",
            string_value=str(config_vars.plugins),
            parameter_name=f"/{config_vars.project}/{config_vars.stage}/pipeline_plugins",
        )

        Aspects.of(self).add(AwsSolutionsChecks(log_ignores=True))



"""Opinionated CDK construct to create S3 bucket.

Security parameters are set by default
"""

import aws_cdk as cdk
import aws_cdk.aws_kms as kms
import aws_cdk.aws_s3 as s3

from constructs import Construct


class S3Bucket(Construct):
    """Create opinionated S3 bucket including restricted public access,
    enforced encryption in transit as well as in rest, versioning and lifecycle
    rules."""

    def __init__(self, scope: Construct, id: str):  # noqa: A002
        super().__init__(scope, id)

    def create_bucket(
        self,
        bucket_name: str,
        encryption: s3.BucketEncryption,
        kms_key: kms.IKey | None = None,
        server_access_logs_bucket: s3.IBucket | None = None,
        enforce_ssl: bool = True,  # noqa: FBT001, FBT002
        **kwargs,
    ) -> s3.Bucket | s3.IBucket:
        """Creates an S3 bucket with opinionated security settings.

        Parameters:

        - bucket_name: Name of the S3 bucket.

        - encryption: Encryption type like S3_MANAGED.

        - kms_key: Optional KMS key for encryption.

        - server_access_logs_bucket: Bucket to store access logs.

        - enforce_ssl: Enforce SSL for all communications.

        - kwargs: Other options like events, log prefix.

        Returns: The S3 Bucket object.

        It sets:

        - Auto object delete on bucket removal
        - Blocks all public access
        - Enables default encryption
        - Versioning and lifecycle rules
        - Secure SSL policy

        It can send logs to another bucket and integrate with EventBridge.
        """

        bucket = s3.Bucket(
            self,
            id=bucket_name,
            auto_delete_objects=True,
            block_public_access=s3.BlockPublicAccess(
                block_public_acls=True,
                block_public_policy=True,
                ignore_public_acls=True,
                restrict_public_buckets=True,
            ),
            bucket_name=bucket_name,
            encryption=encryption,
            encryption_key=kms_key,
            enforce_ssl=bool(enforce_ssl),
            minimum_tls_version=1.2,
            event_bridge_enabled=bool(kwargs.get("event_bridge_enabled")),
            lifecycle_rules=[
                s3.LifecycleRule(
                    enabled=True,
                    noncurrent_version_expiration=cdk.Duration.days(amount=1),
                ),
                s3.LifecycleRule(
                    enabled=True,
                    expired_object_delete_marker=True,
                    abort_incomplete_multipart_upload_after=cdk.Duration.days(7),
                ),
            ],
            public_read_access=False,
            removal_policy=cdk.RemovalPolicy.DESTROY,
            server_access_logs_prefix=kwargs.get("server_access_logs_prefix"),
            server_access_logs_bucket=server_access_logs_bucket,
            versioned=True,
        )

        return bucket



"""Example code for ECR cdk stack."""

from aws_cdk import Aspects, Stack
from cdk_nag import AwsSolutionsChecks, HIPAASecurityChecks, NIST80053R5Checks, PCIDSS321Checks
from constructs import Construct

from cdk_opinionated_constructs.ecr import ECR


class TestECRStack(Stack):
    """TestECRStack defines a CDK stack that creates an ECR repository.

    It creates an ECR construct and configures an ECR repository with provided
    settings.

    It validates the stack against the following checklists:

    - AWS Solutions
    - NIST 800-53 R5
    - PCI DSS 3.2.1
    - HIPAA Security

    Parameters:

    - scope: The CDK scope constructing this stack.
    - construct_id: ID for the stack construct.
    - **kwargs: Additional stack options.
    """

    def __init__(self, scope: Construct, construct_id: str, **kwargs) -> None:
        super().__init__(scope, construct_id, **kwargs)
        ecr_construct = ECR(self, id="ecr_construct")
        ecr_construct.repository(repository_name="repository_name", removal_policy="destroy", max_image_age=90)

        # Validate stack against AWS Solutions checklist
        Aspects.of(self).add(AwsSolutionsChecks(log_ignores=True))
        Aspects.of(self).add(NIST80053R5Checks(log_ignores=True))
        Aspects.of(self).add(PCIDSS321Checks(log_ignores=True))
        Aspects.of(self).add(HIPAASecurityChecks(log_ignores=True))



"""Test CDK template."""

import os

from test.stacks.sns_stack import TestSNSStack

import aws_cdk as cdk
import pytest

from aws_cdk.assertions import Template

CDK_ENV = cdk.Environment(account=os.environ["CDK_DEFAULT_ACCOUNT"], region=os.environ["CDK_DEFAULT_REGION"])


@pytest.fixture(autouse=True)
def stack_template() -> Template:
    """The stack_template fixture generates a CDK stack template for testing.

    It creates a CDK app, instantiates the TestSNSStack, and returns a template
    generated from the stack.

    Parameters:

    - None

    Returns:

    - Template: The CDK stack template generated from TestSNSStack.
    """

    app = cdk.App()
    stack = TestSNSStack(app, "TestSNSStack", env=CDK_ENV)
    return Template.from_stack(stack)


def test_sns_topic_existence(stack_template):
    """test_sns_topic_existence tests that the stack template contains one SNS
    Topic resource.

    It asserts that the template contains exactly one
    "AWS::SNS::Topic" resource.

    Parameters:

    - stack_template: The CDK stack template to test against.
    """

    stack_template.resource_count_is("AWS::SNS::Topic", 1)



import aws_cdk as cdk
import aws_cdk.aws_ssm as ssm

from aws_cdk import Aspects
from cdk_nag import AwsSolutionsChecks
from cdk_opinionated_constructs.schemas.configuration_vars import ConfigurationVars
from cdk_opinionated_constructs.stacks import reduce_items_number, set_ssm_parameter_tier_type
from cdk_opinionated_constructs.utils import load_properties
from constructs import Construct


class IntegrationTestsStack(cdk.Stack):
    """IntegrationTestsStack defines a CDK stack for integration tests
    configuration.

    It loads configuration files from the cdk/config directory into a dict.
    It creates a ConfigurationVars object from the combined props and config.

    It stores the ConfigurationVars as a StringParameter in SSM, with a name built from
    the config values.

    It validates the stack against the AWS Solutions checklist.

    Parameters:

    - scope: The CDK scope constructing this stack.
    - construct_id: ID for the stack construct.
    - env: The CDK environment.
    - props: Configuration properties passed to the stack.
    - **kwargs: Additional stack options.
    """

    def __init__(self, scope: Construct, construct_id: str, env, props, **kwargs) -> None:
        super().__init__(scope, construct_id, env=env, **kwargs)
        config_vars = ConfigurationVars(**props)

        props_env = load_properties(stage=config_vars.stage)
        ssm_parameter_value = reduce_items_number(values=props_env)

        ssm.StringParameter(
            self,
            id="config_file",
            string_value=str(ssm_parameter_value),
            parameter_name=f"/{config_vars.project}/{config_vars.stage}/integration/tests/config",
            tier=set_ssm_parameter_tier_type(character_number=len(str(ssm_parameter_value))),
        )

        Aspects.of(self).add(AwsSolutionsChecks(log_ignores=True))



import aws_cdk as cdk

from cdk_opinionated_constructs.stacks.governance_stack import GovernanceStack
from cdk_opinionated_constructs.stacks.notifications_stack import NotificationsStack
from constructs import Construct


class NotificationsStage(cdk.Stage):
    def __init__(self, scope: Construct, construct_id: str, env: cdk.Environment, props: dict, **kwargs) -> None:
        """Constructs the Notifications and Governance stacks.

        The NotificationsStack constructs resources related to
        notifications, such as SNS topics and subscriptions.

        The GovernanceStack constructs resources related to governance,
        such as Config and GuardDuty.
        """
        super().__init__(scope, construct_id, env=env, **kwargs)

        NotificationsStack(
            self,
            construct_id="notifications-stack",
            env=env,
            props=props,
        )

        GovernanceStack(
            self,
            construct_id="governance-stack",
            env=env,
            props=props,
        )



"""Test CDK template."""

import os

from test.stacks.lmb_stack import TestAWSPythonLambdaFunctionStack

import aws_cdk as cdk
import pytest

from aws_cdk.assertions import Template

CDK_ENV = cdk.Environment(account=os.environ["CDK_DEFAULT_ACCOUNT"], region=os.environ["CDK_DEFAULT_REGION"])


@pytest.fixture(autouse=True)
def stack_template() -> Template:
    """The stack_template fixture generates a CDK stack template for testing.

    It creates a CDK app, instantiates the TestAWSPythonLambdaFunctionStack,
    and returns a template generated from the stack.

    Parameters:

    - None

    Returns:

    - Template: The CDK stack template generated from TestAWSPythonLambdaFunctionStack.
    """

    app = cdk.App()
    stack = TestAWSPythonLambdaFunctionStack(
        app,
        "TestAWSLambdaFunctionStack",
        env=CDK_ENV,
        props={"stage": "dev", "project": "test_project", "service_name": "example_lambda_function"},
    )
    return Template.from_stack(stack)


def test_aws_lambda_function_existence(stack_template):
    """test_aws_lambda_function_existence tests that the stack template
    contains two AWS Lambda Function resources.

    It asserts that the template contains exactly two
    "AWS::Lambda::Function" resources.

    Parameters:

    - stack_template: The CDK stack template to test against.
    """

    stack_template.resource_count_is("AWS::Lambda::Function", 2)


def test_signer_profile_existence(stack_template):
    """test_signer_profile_existence tests that the stack template contains one
    AWS Signer SigningProfile resource.

    It asserts that the template contains exactly one
    "AWS::Signer::SigningProfile" resource.

    Parameters:

    - stack_template: The CDK stack template to test against.
    """

    stack_template.resource_count_is("AWS::Signer::SigningProfile", 1)



"""A setuptools based setup module.

See:
https://packaging.python.org/en/latest/distributing.html
"""

from setuptools import find_packages, setup

setup(
    name="cdk-opinionated-constructs",
    version="3.13.0",
    description="AWS CDK constructs come without added security configurations.",
    long_description="The idea behind this project is to create secured constructs from the start. \n"
    "Supported constructs: ALB, ECR, LMB, NLB, S3, SNS, WAF, RDS",
    license="MIT",
    package_dir={"": "."},
    packages=find_packages(where="."),
    install_requires=[
        "aws-cdk-lib>=2.167.0",
        "constructs>=10.4.2",
        "cdk-monitoring-constructs>=6.0.0",
        "pydantic>=2.9.0",
        "pydantic-core>=2.24.0",
        "pyyaml>=6.0.0",
        "tenacity>=8.0.1",
    ],
    python_requires=">=3.11",
)



import aws_cdk.aws_chatbot as chatbot
import aws_cdk.aws_events as events
import aws_cdk.aws_events_targets as events_targets
import aws_cdk.aws_iam as iam
import aws_cdk.aws_sns as sns
import aws_cdk.aws_sns_subscriptions as sns_subscriptions
import aws_cdk.aws_ssm as ssm

from aws_cdk.aws_codestarnotifications import DetailType, NotificationRule
from cdk.schemas.configuration_vars import PipelineVars
from cdk_nag import NagPackSuppression
from cdk_opinionated_constructs.utils import apply_tags


def count_characters_number(values: dict[list, dict] | dict[str, str]) -> int:
    """Counts the number of characters in the values.

    Parameters:
        - values: The environment configuration properties.

    Returns:
      - The number of characters.
    """
    return len(str(values))


def reduce_items_number(
    values: dict[list, dict], standard_character_number: int = 4096
) -> dict[list, dict] | dict[str, str]:
    """Counts the number of characters in the values.

    Parameters:
        - values: The environment configuration properties.
        - standard_character_number: The standard number of characters.

    Returns:
      - A Dict with the reduced number of items.
    """

    char_count = count_characters_number(values)

    if char_count <= standard_character_number:
        return values

    return {"dummy": "value"}


def set_ssm_parameter_tier_type(*, character_number: int) -> ssm.ParameterTier:
    """Sets the tier type of the parameter based on the total characters of the
    key and value.

    Parameters:
        - values: The environment configuration properties.

    Returns:
      - The tier type.
    """

    tier_type = ssm.ParameterTier.STANDARD
    if 4069 <= character_number <= 8192:
        tier_type = ssm.ParameterTier.ADVANCED
    return tier_type


def notifications_topic(self, pipeline_vars: PipelineVars) -> sns.Topic:
    """Creates an SNS topic for pipeline notifications.

    Parameters:

    pipeline_vars (PipelineVars): A model containing pipeline configuration values like notification email.

    Functionality:

    1. create an SNS topic called 'notifications_topic' to be used for notifications.

    2. add a subscription to the topic using the email address from pipeline_vars.

    3. add metadata to suppress cdk_nag warnings about encryption.

    4. return the created SNS topic.
    """
    notifications_sns_topic = sns.Topic(self, "notifications_topic", display_name="CodePipeline notifications")
    notifications_sns_topic.add_subscription(
        topic_subscription=sns_subscriptions.EmailSubscription(email_address=pipeline_vars.ci_cd_notification_email)
    )

    # Warning suppression for cdk_nag
    notifications_sns_topic_cfn = notifications_sns_topic.node.default_child
    notifications_sns_topic_cfn.add_metadata(  # type: ignore
        "cdk_nag",
        {
            "rules_to_suppress": [
                NagPackSuppression(id="AwsSolutions-SNS2", reason="Notifications stack, doesn't require encryption"),
            ]
        },
    )
    return notifications_sns_topic


def pipeline_notifications(self, sns_topic: sns.ITopic) -> None:
    """Configures notifications for pipeline events.

    Parameters:

    sns_topic (sns.ITopic): The SNS topic to send notifications to.

    Functionality:

    1. add resource policy to an SNS topic to allow CodeStar Notifications to publish.

    2. create a NotificationRule to send notifications on pipeline events:
       - Pipeline execution failed
       - Action execution failed
       - Stage execution failed
       - Manual approval failed
       - Manual approval needed

    the notification rule will send messages to the provided SNS topic.
    """
    sns_topic.add_to_resource_policy(
        iam.PolicyStatement(
            effect=iam.Effect.ALLOW,
            principals=[iam.ServicePrincipal(service="codestar-notifications.amazonaws.com")],
            actions=["SNS:Publish"],
            resources=[sns_topic.topic_arn],
        )
    )

    # The CodePipeline notifications available rules:
    # https://docs.aws.amazon.com/dtconsole/latest/userguide/concepts.html#events-ref-repositories
    NotificationRule(
        self,
        "codepipeline_notifications",
        detail_type=DetailType.FULL,
        events=[
            "codepipeline-pipeline-action-execution-canceled",
            "codepipeline-pipeline-action-execution-failed",
            "codepipeline-pipeline-action-execution-started",
            "codepipeline-pipeline-action-execution-succeeded",
            "codepipeline-pipeline-manual-approval-failed",
            "codepipeline-pipeline-manual-approval-needed",
            "codepipeline-pipeline-manual-approval-succeeded",
            "codepipeline-pipeline-pipeline-execution-canceled",
            "codepipeline-pipeline-pipeline-execution-failed",
            "codepipeline-pipeline-pipeline-execution-resumed",
            "codepipeline-pipeline-pipeline-execution-started",
            "codepipeline-pipeline-pipeline-execution-succeeded",
            "codepipeline-pipeline-pipeline-execution-superseded",
            "codepipeline-pipeline-stage-execution-canceled",
            "codepipeline-pipeline-stage-execution-failed",
            "codepipeline-pipeline-stage-execution-resumed",
            "codepipeline-pipeline-stage-execution-started",
            "codepipeline-pipeline-stage-execution-succeeded",
        ],
        source=self.codepipeline.pipeline,
        targets=[sns_topic],
    )


def pipeline_email_notifications(sns_topic: sns.Topic) -> None:
    """Configures email notifications for the pipeline SNS topic.

    Parameters:

    sns_topic (sns.Topic): The SNS topic to configure notifications for.

    Functionality:

    - adds a resource policy to the SNS topic to allow CodeStar Notifications
    to publish messages to it.

    - this enables email notifications to be sent on pipeline events.
    """
    sns_topic.add_to_resource_policy(
        iam.PolicyStatement(
            effect=iam.Effect.ALLOW,
            principals=[iam.ServicePrincipal(service="codestar-notifications.amazonaws.com")],
            actions=["SNS:Publish"],
            resources=[sns_topic.topic_arn],
        )
    )


def pipeline_trigger(self, pipeline_vars: PipelineVars, props: dict, schedule: events.Schedule):
    """Creates a scheduled rule to trigger the pipeline.

    Parameters:

    pipeline_vars (PipelineVars): A model containing pipeline configuration values.
    props (dict): A dictionary of configuration properties.
    schedule (events.Schedule): The scheduled interval to trigger the pipeline.
    self:

    Functionality:
    1. create a Rule with the provided schedule, enabled status, name, and description.
    2. add the pipeline as a target for the rule.
    this will trigger the pipeline on the schedule.
    3. apply tags to the rule based on props.
    """

    # Auto triggers the pipeline every day to ensure pipeline validation
    trigger = events.Rule(
        self,
        id="daily_release",
        description="Auto trigger the pipeline every day to ensure pipeline validation",
        enabled=True,
        rule_name=f"{pipeline_vars.project}-scheduled-release",
        schedule=schedule,
    )
    trigger.add_target(events_targets.CodePipeline(self.codepipeline.pipeline))  # type: ignore
    apply_tags(props=props, resource=trigger)  # type: ignore


def _configure_pipeline_email_notifications(sns_topic: sns.Topic) -> None:
    """Configures email notifications for the pipeline SNS topic.

    This is an internal helper function to avoid redundant code.

    Parameters:

    sns_topic (sns.Topic): The SNS topic to configure notifications for.
    """
    sns_topic.add_to_resource_policy(
        iam.PolicyStatement(
            effect=iam.Effect.ALLOW,
            principals=[iam.ServicePrincipal(service="codestar-notifications.amazonaws.com")],
            actions=["SNS:Publish"],
            resources=[sns_topic.topic_arn],
        )
    )


def _configure_pipeline_slack_notifications(
    scope, notifications_sns_topic: sns.Topic | sns.ITopic, pipeline_vars: PipelineVars
) -> None:
    """Configures Slack notifications for the pipeline.

    This is an internal helper function to encapsulate Slack-specific logic.

    Parameters:
    scope: The scope for the chatbot construct.
    notifications_sns_topic (sns.Topic | sns.ITopic): The SNS topic to send notifications to.
    pipeline_vars (PipelineVars): Pipeline variables containing Slack configuration.
    """
    if pipeline_vars.slack_workspace_id and pipeline_vars.slack_ci_cd_channel_id:
        chatbot.SlackChannelConfiguration(
            scope,
            "chatbot",
            slack_channel_configuration_name=pipeline_vars.project,
            notification_topics=[notifications_sns_topic],
            slack_workspace_id=pipeline_vars.slack_workspace_id,
            slack_channel_id=pipeline_vars.slack_ci_cd_channel_id,
        )


def _configure_pipeline_ms_teams_notifications(
    scope, notifications_sns_topic: sns.Topic | sns.ITopic, pipeline_vars: PipelineVars
) -> None:
    """Configures MS Teams notifications for the pipeline.

    This is an internal helper function to encapsulate MS Teams-specific logic.

    Parameters:
    scope: The scope for the chatbot construct.
    notifications_sns_topic (sns.Topic | sns.ITopic): The SNS topic to send notifications to.
    pipeline_vars (PipelineVars): Pipeline variables containing MS Teams configuration.
    """
    if pipeline_vars.ms_teams_team_id and pipeline_vars.ms_teams_ci_cd_channel_id:
        chatbot.CfnMicrosoftTeamsChannelConfiguration(
            scope,
            "ci-cd-ms-teams-chatbot",
            configuration_name=f"{pipeline_vars.project}-ci-cd",
            sns_topic_arns=[notifications_sns_topic.topic_arn],
            team_id=pipeline_vars.ms_teams_team_id,
            teams_channel_id=pipeline_vars.ms_teams_ci_cd_channel_id,
            teams_tenant_id=pipeline_vars.ms_teams_tenant_id,
        )


def create_pipeline_notifications(
    scope,
    notifications_sns_topic: sns.Topic | sns.ITopic,
    pipeline_vars: PipelineVars,
    use_chatbot: bool = True,  # noqa: FBT001, FBT002
):
    """Configures notifications for the pipeline.

    This function orchestrates the configuration of different notification channels.

    Parameters:
    scope: The scope for child constructs.
    notifications_sns_topic (sns.Topic | sns.ITopic): The SNS topic to send notifications to.
    pipeline_vars (PipelineVars): Pipeline variables containing notification configurations.
    use_chatbot (bool): Flag to enable/disable chatbot integrations.
    """

    if pipeline_vars.ci_cd_notification_email:
        _configure_pipeline_email_notifications(sns_topic=notifications_sns_topic)
    if pipeline_vars.slack_ci_cd_channel_id:
        pipeline_notifications(scope, sns_topic=notifications_sns_topic)

    if use_chatbot:
        _configure_pipeline_slack_notifications(scope, notifications_sns_topic, pipeline_vars)
        _configure_pipeline_ms_teams_notifications(scope, notifications_sns_topic, pipeline_vars)



import aws_cdk as cdk
import aws_cdk.aws_sns as sns
import aws_cdk.aws_ssm as ssm

from aws_cdk import Aspects
from aws_cdk.aws_budgets import CfnBudget as Budget
from cdk_nag import AwsSolutionsChecks, NagPackSuppression, NagSuppressions
from cdk_opinionated_constructs.schemas.configuration_vars import ConfigurationVars, GovernanceVars
from cdk_opinionated_constructs.utils import load_properties
from constructs import Construct


class GovernanceStack(cdk.Stack):
    def __init__(self, scope: Construct, construct_id: str, env, props, **kwargs) -> None:
        """Initializes the GovernanceStack.

        This stack is responsible for setting up AWS governance-related resources. It includes:
        - Importing an SNS topic from an ARN stored in SSM Parameter Store.
        - Adding a resource policy to the SNS topic to allow AWS Budgets to publish to it.
        - Suppressing a specific cdk-nag rule for the SNS topic.
        - Creating SNS subscribers for budget alarms with daily and monthly periods.
        - Attaching AWS Solutions checks to the stack for compliance monitoring.

        Args:
            scope (Construct): The parent construct.
            construct_id (str): The construct's unique identifier.
            env: The AWS environment (account/region) where this stack will be deployed.
            props: The properties for configuring the stack.
            **kwargs: Additional keyword arguments.
        """
        super().__init__(scope, construct_id, env=env, **kwargs)
        config_vars = ConfigurationVars(**props)

        props_env = load_properties(stage=config_vars.stage)

        governance_vars = GovernanceVars(**props_env)

        bill_sns_topic = sns.Topic.from_topic_arn(
            self,
            id="imported_sns_topic",
            topic_arn=ssm.StringParameter.value_for_string_parameter(
                self,
                parameter_name=f"/{config_vars.project}/{config_vars.stage}/topic/alarm/arn",
            ),
        )

        NagSuppressions.add_resource_suppressions(
            bill_sns_topic,
            [NagPackSuppression(id="AwsSolutions-SNS2", reason="Notifications stack, doesn't require encryption")],
        )

        bill_sns_subscribers = Budget.SubscriberProperty(address=bill_sns_topic.topic_arn, subscription_type="SNS")

        if governance_vars.budget_limit_monthly and governance_vars.tags.awsApplication:
            monthly_budget_limit = governance_vars.budget_limit_monthly
            self.budget_alarms(
                sns_subscribers=bill_sns_subscribers,
                period="DAILY",
                config_vars=config_vars,
                governance_vars=governance_vars,
                monthly_budget_limit=monthly_budget_limit,
            )
            self.budget_alarms(
                sns_subscribers=bill_sns_subscribers,
                period="MONTHLY",
                config_vars=config_vars,
                governance_vars=governance_vars,
                monthly_budget_limit=monthly_budget_limit,
            )

        Aspects.of(self).add(AwsSolutionsChecks(log_ignores=True))

    def budget_alarms(
        self,
        *,
        sns_subscribers: Budget.SubscriberProperty,
        period: str,
        config_vars: ConfigurationVars,
        governance_vars: GovernanceVars,
        monthly_budget_limit: float,
        budget_threshold: int = 95,
    ) -> None:
        """Creates budget alarms for both forecasted and actual spend.

        This method sets up AWS Budgets with alarms for when the forecasted or actual
        costs exceed the specified threshold. It supports different periods (e.g., DAILY, MONTHLY)
        and allows for setting a monthly budget limit. Notifications for breaching the budget
        are sent to the provided SNS subscribers.

        Parameters:
        - sns_subscribers: An instance of Budget.SubscriberProperty to receive notifications.
        - period: The period for the budget, e.g., 'DAILY' or 'MONTHLY'.
        - config_vars: Configuration variables containing project and stage names.
        - governance_vars: Governance-related variables, including tags.
        - monthly_budget_limit: The monthly budget limit in USD.
        - budget_threshold: The threshold percentage for triggering the alarm (default is 95%).

        Returns:
        None
        """
        amount = {
            "DAILY": monthly_budget_limit / 30,
            "MONTHLY": monthly_budget_limit,
        }

        # Forecasted spend, DAILY budget only supports a notification type as ACTUAL
        if period != "DAILY":
            budget_name = f"{config_vars.project}-{config_vars.stage}-forecasted-{period.lower()}"
            Budget(
                self,
                id=budget_name,
                budget=Budget.BudgetDataProperty(
                    budget_limit=Budget.SpendProperty(amount=amount[period], unit="USD"),
                    budget_name=budget_name,
                    budget_type="COST",
                    cost_filters={"TagKeyValue": [f"user:awsApplication${governance_vars.tags.awsApplication}"]},
                    time_unit=period,
                ),
                notifications_with_subscribers=[
                    Budget.NotificationWithSubscribersProperty(
                        notification=Budget.NotificationProperty(
                            threshold=budget_threshold,
                            notification_type="FORECASTED",
                            comparison_operator="GREATER_THAN",
                        ),
                        subscribers=[sns_subscribers],
                    )
                ],
            )
        # Current spend
        budget_name = f"{config_vars.project}-{config_vars.stage}-current-{period.lower()}"
        Budget(
            self,
            id=budget_name,
            budget=Budget.BudgetDataProperty(
                budget_limit=Budget.SpendProperty(amount=amount[period], unit="USD"),
                budget_name=budget_name,
                budget_type="COST",
                cost_filters={"TagKeyValue": [f"user:awsApplication${governance_vars.tags.awsApplication}"]},
                time_unit=period,
            ),
            notifications_with_subscribers=[
                Budget.NotificationWithSubscribersProperty(
                    notification=Budget.NotificationProperty(
                        threshold=budget_threshold, notification_type="ACTUAL", comparison_operator="GREATER_THAN"
                    ),
                    subscribers=[sns_subscribers],
                )
            ],
        )
